/*************************************************************
 *
 *  Copyright (c) 2017 The MathJax Consortium
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */


/**
 * @fileoverview The TexParser. Implements the basic parsing functionality and
 *     administers the global stack and tree objects.
 *
 * @author v.sorge@mathjax.org (Volker Sorge)
 */

import * as sm from './SymbolMap.js';
import MapHandler from './MapHandler.js';
import {HandlerType, Configuration, ParseInput, ParseResult, ParseMethod} from './Types.js';
import {BaseMappings} from './BaseMappings.js';
import {AmsMappings} from './AmsMappings.js';
import {AmsSymbols} from './AmsSymbols.js';
import {ParseMethods} from './ParseMethods.js';

// From OLD Parser
import {TreeHelper} from './TreeHelper.js';
import {MmlNode} from '../../core/MmlTree/MmlNode.js';
import {MmlMo} from '../../core/MmlTree/MmlNodes/mo.js';
import {StackItem, StopItem, EnvList} from './StackItem.js';
import Stack from './Stack.js';
import {Symbol} from './Symbol.js';
import TexError from './TexError.js';


/**
 * The main Tex Parser class.
 */
export default class TexParser {

  public macroCount: number = 0;
  public secondPass: MmlMo[] = [];

  private input: string = '';
  private remainder: string = '';
  private configurations: Map<HandlerType, SubHandler> = new Map();
  
  // From OLD Parser
  NBSP = '\u00A0'; 
  string: string;
  i: number = 0;
  remap: sm.CharacterMap = MapHandler.getInstance().getMap('remap') as sm.CharacterMap;
  stack: Stack;
  
  
  /**
   * @constructor
   */
  constructor(str: string, env: EnvList) {
    // TODO: Move this into a configuration object.
    this.configure(BaseMappings.CONFIGURATION);
    this.append(AmsMappings.CONFIGURATION);
    this.append(AmsSymbols.CONFIGURATION);
    this.setup(ParseMethods as any);

    this.string = str;
    let ENV: EnvList;
    if (env) {
      ENV = {};
      for (let id in env) {
        if (env.hasOwnProperty(id)) {
          ENV[id] = env[id]}
      }
    }
    this.stack = new Stack(ENV, !!env);
    // configurations.forEach(this.NewParser.append.bind(this.NewParser));
    this.Parse();
    this.Push(new StopItem());
  }


  /**
   * The main parsing method.
   * @param {string} tex The TeX string.
   * @return {MmlNode} The parsing result.
   */
  // TODO (VS): This will eventually become the actual parsing method.
  public process(tex: string): void { }


  /**
   * Sets up parsing methods etc. from legacy code.
   * @param {Record.<string, ParseMethod} env The MathJax legacy object.
   */
  // TODO (VS): Temporary for setting up parsing in SymbolMaps.
  public setup(env: Record<string, ParseMethod>) {
    const maps = MapHandler.getInstance().allMaps();
    for (let i = 0, map; map = maps[i]; i++) {
      if (map instanceof sm.CharacterMap ||
          map instanceof sm.RegExpMap ||
          map instanceof sm.EnvironmentMap) {
        try {
          let parser = map.parser(null);
          if (typeof parser === 'string') {
            map.parser = env[parser];
          }
        } catch (e) {}
      }
      if (map instanceof sm.MacroMap) {
        map.functionMap = env;
      }
    }
    this.fallback('character', env['Other']);
    this.fallback('macro', env['csUndefined']);
    this.fallback('environment', env['envUndefined']);
  }


  /**
   * Sets a new configuration for the map handler.
   * @param {{character: Array.<string>,
   *          delimiter: Array.<string>,
   *          macro: Array.<string>,
   *          environment: Array.<string>}} configuration A setting for the
   *    map handler.
   */
  public configure(config: Configuration): void {
    for (const key of Object.keys(config)) {
      let name = key as HandlerType;
      this.configurations.set(name, new SubHandler(config[name] || []));
    }
  }


  /**
   * Appends configurations to the current map handlers.
   * @param {{character: Array.<string>,
   *          delimiter: Array.<string>,
   *          macro: Array.<string>,
   *          environment: Array.<string>}} configuration A setting for the
   *    map handler.
   */
  public append(config: Configuration): void {
    for (const key of Object.keys(config)) {
      let name = key as HandlerType;
      for (const map of config[name]) {
        this.configurations.get(name).add(map);
      }
    }
  }


  /**
   * Parses the input with the specified kind of map.
   * @param {HandlerType} kind Configuration name.
   * @param {ParseInput} input Input to be parsed.
   * @return {ParseResult} The output of the parsing function.
   */
  public parse(kind: HandlerType, input: ParseInput): ParseResult {
    return this.configurations.get(kind).parse(input);
  }


  /**
   * Maps a symbol to its "parse value" if it exists.
   *
   * @param {HandlerType} kind Configuration name.
   * @param {string} symbol The symbol to parse.
   * @return {T} A boolean, Character, or Macro.
   */
  public lookup(kind: HandlerType, symbol: string) {
    return this.configurations.get(kind).lookup(symbol);
  }


  /**
   * Maps a symbol to its "parse value" if it exists.
   *
   * @param {HandlerType} kind Configuration name.
   * @param {string} symbol The symbol to parse.
   * @return {T} A boolean, Character, or Macro.
   */
  public fallback(kind: HandlerType, method: ParseMethod) {
    return this.configurations.get(kind).fallback(method);
  }

  /**
   * Checks if a symbol is contained in one of the symbol mappings of the
   * specified kind.
   *
   * @param {string} symbol The symbol to parse.
   * @return {boolean} True if the symbol is contained in the given types of
   *     symbol mapping.
   */
  public contains(kind: HandlerType, symbol: string): boolean {
    return this.configurations.get(kind).contains(symbol);
  }


  /**
   * @override
   */
  public toString(): string {
    let str = '';
    for (const config of Array.from(this.configurations.keys())) {
      str += config + ': ' +
        this.configurations.get(config as HandlerType) + '\n';
    }
    return str;
  }

  // From OLD Parser
  public Parse() {
    TreeHelper.printMethod('Parse (Old Parser Object)');
    let c, n;
    while (this.i < this.string.length) {
      c = this.string.charAt(this.i++); n = c.charCodeAt(0);
      if (n >= 0xD800 && n < 0xDC00) {
        c += this.string.charAt(this.i++);
      }
      this.parse('character', [c, this]);
    }
  }

  public Push(arg: StackItem|MmlNode) {
    TreeHelper.printMethod('Push (Old Parser Object)');
    this.stack.Push(arg);
  }

  public PushAll(args: (StackItem|MmlNode)[]) {
    TreeHelper.printMethod('PushAll (Old Parser Object)');
    for (let i = 0, m = args.length; i < m; i++) {
      this.stack.Push(args[i]);
    }
  }


  /**
   * @return {MmlNode} The internal Mathml structure. 
   */
  public mml(): MmlNode {
    TreeHelper.printMethod('mml (Old Parser Object)');
    if (!this.stack.Top().hasType('mml')) {
      return null;
    }
    let node = this.stack.Top().data[0];
    // Makes sure TeXclasses are properly set, so none is null.
    node.setTeXclass(null);
    return node;
  }

  // VS: Forget this for now!
  public mmlToken(token: MmlNode): MmlNode {
    return token;
  } // used by boldsymbol extension


  /************************************************************************/
  /*
   *   String handling routines
   */

  /**
   *  Convert delimiter to character
   */
  public convertDelimiter(c: string): string {
    TreeHelper.printMethod('convertDelimiter (Old Parser Object)');
    const symbol = this.lookup('delimiter', c) as Symbol;
    return symbol ? symbol.char : null;
  }

  /**
   *  Trim spaces from a string
   */
  // static
  public trimSpaces(text: string): string {
    TreeHelper.printMethod('trimSpaces (Old Parser Object)');
    if (typeof(text) != 'string') {return text}
    let TEXT = text.replace(/^\s+|\s+$/g,'');
    if (TEXT.match(/\\$/) && text.match(/ $/)) TEXT += ' ';
    return TEXT;
  }

  /**
   *   Check if the next character is a space
   */
  public nextIsSpace() {
    TreeHelper.printMethod('nextIsSpace (Old Parser Object)');
    return this.string.charAt(this.i).match(/\s/);
  }

  
  /**
   *  Get the next non-space character
   */
  public GetNext(): string {
    TreeHelper.printMethod('GetNext (Old Parser Object)');
    while (this.nextIsSpace()) {this.i++}
    return this.string.charAt(this.i);
  }
  
  /**
   *  Get and return a control-sequence name
   */
  // TODO: The argument is given once in GetDelimiter, but never used!
  //       Check with Davide!
  public GetCS(name?: string) {
    TreeHelper.printMethod('GetCS (Old Parser Object)');
    let CS = this.string.slice(this.i).match(/^([a-z]+|.) ?/i);
    if (CS) {
      this.i += CS[1].length;
      return CS[1]
    } else {
      this.i++;
      return ' ';
    }
  }

  /**
   *  Get and return a TeX argument (either a single character or control sequence,
   *  or the contents of the next set of braces).
   */
  public GetArgument(name: string, noneOK?: boolean) {
    TreeHelper.printMethod('GetArgument (Old Parser Object)');
    switch (this.GetNext()) {
    case '':
      if (!noneOK) {throw new TexError(['MissingArgFor','Missing argument for %1',name])}
      return null;
    case '}':
      if (!noneOK) {
        throw new TexError(['ExtraCloseMissingOpen',
                            'Extra close brace or missing open brace']);
      }
      return null;
    case '\\':
      this.i++; return '\\'+this.GetCS();
    case '{':
      let j = ++this.i, parens = 1;
      while (this.i < this.string.length) {
        switch (this.string.charAt(this.i++)) {
        case '\\':  this.i++; break;
        case '{':   parens++; break;
        case '}':
          if (--parens == 0) {return this.string.slice(j,this.i-1)}
          break;
        }
      }
      throw new TexError(['MissingCloseBrace','Missing close brace']);
    }        
    return this.string.charAt(this.i++);
  }

  
  /**
   *  Get an optional LaTeX argument in brackets
   */
  public GetBrackets(name: string, def?: string): string {
    TreeHelper.printMethod('GetBrackets (Old Parser Object)');
    if (this.GetNext() != '[') {return def};
    let j = ++this.i, parens = 0;
    while (this.i < this.string.length) {
      switch (this.string.charAt(this.i++)) {
      case '{':   parens++; break;
      case '\\':  this.i++; break;
      case '}':
        if (parens-- <= 0) {
          throw new TexError(['ExtraCloseLooking',
                              'Extra close brace while looking for %1','\']\'']);
        }
        break;   
      case ']':
        if (parens == 0) {
          return this.string.slice(j,this.i-1)
        }
        break;
      }
    }
    throw new TexError(['MissingCloseBracket',
                        'Could not find closing \']\' for argument to %1', name]);
  }
  
  /**
   *  Get the name of a delimiter (check it in the delimiter list).
   */
  public GetDelimiter(name: string, braceOK?: boolean) {
    TreeHelper.printMethod('GetDelimiter (Old Parser Object)');
    while (this.nextIsSpace()) {this.i++}
    let c = this.string.charAt(this.i); this.i++;
    if (this.i <= this.string.length) {
      if (c == '\\') {
        c += this.GetCS(name);
      } else if (c === '{' && braceOK) {
        this.i--;
        c = this.GetArgument(name);
      }
      if (this.contains('delimiter', c)) {
        return this.convertDelimiter(c);
      }
    }
    throw new TexError(['MissingOrUnrecognizedDelim',
                        'Missing or unrecognized delimiter for %1',name]);
  }

  /**
   *  Get a dimension (including its units).
   */
  public GetDimen(name: string) {
    TreeHelper.printMethod('GetDimen (Old Parser Object)');
    if (this.nextIsSpace()) {this.i++}
    if (this.string.charAt(this.i) == '{') {
      let dimen = this.GetArgument(name);
      if (dimen.match(/^\s*([-+]?([.,]\d+|\d+([.,]\d*)?))\s*(pt|em|ex|mu|px|mm|cm|in|pc)\s*$/))
      {return dimen.replace(/ /g,'').replace(/,/,'.')}
    } else {
      let dimen = this.string.slice(this.i);
      let match = dimen.match(/^\s*(([-+]?([.,]\d+|\d+([.,]\d*)?))\s*(pt|em|ex|mu|px|mm|cm|in|pc)) ?/);
      if (match) {
        this.i += match[0].length;
        return match[1].replace(/ /g,'').replace(/,/,'.');
      }
    }
    throw new TexError(['MissingDimOrUnits',
                        'Missing dimension or its units for %1',name]);
  }

  
  /**
   *  Get everything up to the given control sequence (token)
   */
  public GetUpTo(name: string, token: string) {
    TreeHelper.printMethod('GetUpTo (Old Parser Object)');
    while (this.nextIsSpace()) {this.i++}
    let j = this.i;
    let parens = 0;
    while (this.i < this.string.length) {
      let k = this.i;
      let c = this.string.charAt(this.i++);
      switch (c) {
      case '\\':  c += this.GetCS(); break;
      case '{':   parens++; break;
      case '}':
        if (parens == 0) {
          throw new TexError(['ExtraCloseLooking',
                              'Extra close brace while looking for %1',token])
        }
        parens--;
        break;
      }
      if (parens == 0 && c == token) {return this.string.slice(j,k)}
    }
    throw new TexError(['TokenNotFoundForCommand',
                        'Could not find %1 for %2',token,name]);
  }

  /**
   *  Parse various substrings
   */
  ParseArg(name: string) {
    TreeHelper.printMethod('ParseArg (Old Parser Object)');
    let object = new TexParser(this.GetArgument(name), this.stack.env);
    return object.mml();
  }

  ParseUpTo(name: string, token: string) {
    TreeHelper.printMethod('ParseUpTo (Old Parser Object)');
    return new TexParser(this.GetUpTo(name,token), this.stack.env).mml();
  }

  
  /**
   *  Break up a string into text and math blocks
   */
  InternalMath(text: string, level?: number|string) {
    TreeHelper.printMethod('InternalMath (Old Parser Object)');
    let def = (this.stack.env['font'] ? {mathvariant: this.stack.env['font']} : {});
    let mml: MmlNode[] = [], i = 0, k = 0, c, node, match = '', braces = 0;
    if (text.match(/\\?[${}\\]|\\\(|\\(eq)?ref\s*\{/)) {
      while (i < text.length) {
        c = text.charAt(i++);
        if (c === '$') {
          if (match === '$' && braces === 0) {
            // @test Interspersed Text
            node = TreeHelper.createNode('TeXAtom',
                                         [(new TexParser(text.slice(k,i-1),{})).mml()], {});
            mml.push(node);
            match = ''; k = i;
          } else if (match === '') {
            // @test Interspersed Text
            if (k < i-1) mml.push(this.InternalText(text.slice(k,i-1),def));
            match = '$'; k = i;
          }
        } else if (c === '{' && match !== '') {
          braces++;
        } else if (c === '}') {
          if (match === '}' && braces === 0) {
            TreeHelper.untested(12);
            node = TreeHelper.createNode('TeXAtom', [(new TexParser(text.slice(k,i),{})).mml()], def);
            mml.push(node);
            match = ''; k = i;
          } else if (match !== '') {
            if (braces) braces--;
          }
        } else if (c === '\\') {
          if (match === '' && text.substr(i).match(/^(eq)?ref\s*\{/)) {
            // TODO: Sort this out properly. What exactly does it do?
            let len = ((RegExp as any)['$&'] as string).length;
            if (k < i-1) mml.push(this.InternalText(text.slice(k,i-1),def));
            match = '}'; k = i-1; i += len;
          } else {
            c = text.charAt(i++);
            if (c === '(' && match === '') {
              if (k < i-2) mml.push(this.InternalText(text.slice(k,i-2),def));
              match = ')'; k = i;
            } else if (c === ')' && match === ')' && braces === 0) {
              TreeHelper.untested(13);
              node = TreeHelper.createNode('TeXAtom', [(new TexParser(text.slice(k,i-2),{})).mml()], {});
              mml.push(node);
              match = ''; k = i;
            } else if (c.match(/[${}\\]/) && match === '')  {
              i--; text = text.substr(0,i-1) + text.substr(i); // remove \ from \$, \{, \}, or \\
            }
          }
        }
      }
      if (match !== '') throw new TexError(['MathNotTerminated','Math not terminated in text box']);
    }
    if (k < text.length) mml.push(this.InternalText(text.slice(k),def));
    if (level != null) {
      // @test Label, Fbox, Hbox
      mml = [TreeHelper.createNode('mstyle', mml, {displaystyle:false,scriptlevel:level})];
      // VS: OLD
      // mml = [MML.mstyle.apply(MML,mml).With({displaystyle:false,scriptlevel:level})];
    } else if (mml.length > 1) {
      // @test Interspersed Text
      mml = [TreeHelper.createNode('mrow', mml, {})];
      // VS: OLD
      // mml = [MML.mrow.apply(MML,mml)];
    }
    return mml;
  }

  InternalText(text: string, def: EnvList) {
    // @test Label, Fbox, Hbox
    TreeHelper.printMethod('InternalText (Old Parser Object)');
    text = text.replace(/^\s+/, this.NBSP).replace(/\s+$/, this.NBSP);
    let textNode = TreeHelper.createText(text);
    return TreeHelper.createNode('mtext', [], def, textNode);
    // VS: OLD
    // return MML.mtext(MML.chars(text)).With(def);
  }

  // AMS
  
  /**
   *  Get a delimiter or empty argument
   */
  public GetDelimiterArg(name: string) {
    TreeHelper.printMethod('AMS-GetDelimiterArg (Old Parser Object)');
    let c = this.trimSpaces(this.GetArgument(name));
    if (c == '') return null;
    if (this.contains('delimiter', c)) {
      return c;
    }
    // if (c in TEXDEF.delimiter) return c;
    throw new TexError(['MissingOrUnrecognizedDelim','Missing or unrecognized delimiter for %1',name]);
  }
  
  /**
   *  Get a star following a control sequence name, if any
   */
  public GetStar() {
    TreeHelper.printMethod('AMS-GetStar (Old Parser Object)');
    let star = (this.GetNext() === '*');
    if (star) {this.i++}
    return star;
  }

  public GetRemap(char: string): Symbol {
    return this.remap.lookup(char);
  }
  
}




/**
 * Class of symbol mappings that are active in a configuration.
 */
class SubHandler {

  private _configuration: sm.SymbolMap[] = [];
  private _fallback: ParseMethod = (x => { return null; });

  /**
   * @constructor
   * @param {Array.<string>} maps Names of the maps included in this
   *     configuration.
   */
  constructor(maps: string[]) {
    for (const name of maps) {
      this.add(name);
    }
  }


  /**
   * Sets the default method to call when parsing fails.
   * @param {function(string): ParseResult} method The fallback method.
   */
  public fallback(method: ParseMethod) {
    this._fallback = method;
  }


  /**
   * Adds a symbol map to the configuration if it exists.
   * @param {string} name of the symbol map.
   */
  public add(name: string): void {
    let map = MapHandler.getInstance().getMap(name);
    if (!map) {
      this.warn('Configuration ' + name + ' not found! Omitted.');
      return;
    }
    this._configuration.push(map);
  }


  /**
   * Parses the given input with the first applicable symbol map.
   * @param {ParseInput} input The input for the parser.
   * @return {ParseResult} The output of the parsing function.
   */
  public parse(input: ParseInput): ParseResult {
    // TODO: Can't be done with applicable due to delimiter parsing!
    for (let map of this._configuration) {
      const result = map.parse(input);
      if (result) {
        return result;
      }
    }
    let [symbol, env] = input;
    return this._fallback.bind(env)(env, symbol);
  }


  /**
   * Maps a symbol to its "parse value" if it exists.
   *
   * @param {string} symbol The symbol to parse.
   * @return {T} A boolean, Character, or Macro.
   */
  public lookup<T>(symbol: string): T {
    let map = this.applicable(symbol) as sm.AbstractSymbolMap<T>;
    return map ? map.lookup(symbol) : null;
  }


  /**
   * Checks if a symbol is contained in one of the symbol mappings of this
   * configuration.
   *
   * @param {string} symbol The symbol to parse.
   * @return {boolean} True if the symbol is contained in the mapping.
   */
  public contains(symbol: string): boolean {
    return this.applicable(symbol) ? true : false;
  }


  /**
   * @override
   */
  public toString(): string {
    return this._configuration
      .map(function(x: sm.SymbolMap) {return x.name; })
      .join(', ');
  }


  /**
   * Retrieves the first applicable symbol map in the configuration.
   * @param {string} symbol The symbol to parse.
   * @return {SymbolMap} A map that can parse the symbol.
   */
  private applicable(symbol: string): sm.SymbolMap {
    for (let map of this._configuration) {
      if (map.contains(symbol)) {
        return map;
      }
    }
    return null;
  }


  // TODO: Turn this into a global warning and error functionality
  /**
   * Prints a warning message.
   * @param {string} message The warning.
   */
  private warn(message: string) {
    console.log('TexParser Warning: ' + message);
  }

}
